import logging
import math

logger = logging.getLogger(__name__)


class CrudManager(object):

    def __init__(
        self, provider, service, db_storage=None,
        to_serializable=None,
        from_serializable=None,
    ):
        self.db_storage = db_storage
        self.to_serializable = to_serializable
        self.from_serializable = from_serializable

    def get(self, pk):
        obj = self.db_storage.get(pk)
        return self.to_serializable(obj)

    def list(self, filters=None, order_by=None, offset=None, limit=None):
        results = self.db_storage.list(
            filters=filters, order_by=order_by, offset=offset, limit=limit
        )
        return [self.to_serializable(result) for result in results]

    def page(self, page_size, page_num, filters=None, order_by=None):
        if page_size < 1:
            raise ValueError('Invalid page_size ({})'.format(page_size))
        if page_num < 1:
            raise ValueError('Invalid page_num ({})'.format(page_num))

        offset = page_size * (page_num - 1)
        limit = page_size
        total = self.count(filters=filters)
        num_pages = math.ceil(total / page_size)
        results = self.list(
            filters=filters, order_by=order_by, offset=offset, limit=limit
        )
        return {
            'results': results,
            'num_pages': num_pages,
            'num_results': total,
            'page_num': page_num,
        }

    def count(self, filters=None):
        return self.db_storage.count(filters=filters)

    def update(self, pk, data):
        data = self.from_serializable(data)
        updated_obj = self.db_storage.update(pk, data)
        updated_data = self.to_serializable(updated_obj)
        return updated_data

    def _create_object(self, data):
        data = self.from_serializable(data)
        return self.db_storage.create(data)

    def create(self, data):
        created_obj = self._create_object(data)
        return self.to_serializable(created_obj)

    def delete(self, pk):
        deleted_data = self.get(pk)
        self.db_storage.delete(pk)
        return deleted_data


class CrudManagerWithEvents(CrudManager):

    def __init__(
        self, provider, service,
        event_entity_name=None, dispatcher_accessor=None,
        create_event_name=None, update_event_name=None, delete_event_name=None,
        to_event_serializable=None, **kwargs
    ):
        super(CrudManagerWithEvents, self).__init__(
            provider, service, **kwargs)

        self.event_entity_name = event_entity_name
        self.dispatcher = dispatcher_accessor(service)
        self.to_event_serializable = (
            to_event_serializable or self.to_serializable
        )
        self.create_event_name = create_event_name
        self.update_event_name = update_event_name
        self.delete_event_name = delete_event_name

    def _dispatch_event(self, event_name, object_data, payload=None):
        if event_name:
            payload = payload or {}
            payload.update({self.event_entity_name: object_data})
            self.dispatcher(event_name, payload)
            logger.info('dispatched event: %s', event_name)

    def update(self, pk, data):
        before_obj = self.db_storage.get(pk)
        before_data = self.to_event_serializable(before_obj)

        updated_data = super(CrudManagerWithEvents, self).update(pk, data)

        after_obj = self.db_storage.get(pk)
        after_data = self.to_event_serializable(after_obj)

        if before_data != after_data:
            changed = [
                field for field in sorted(set(before_data).union(after_data))
                if before_data.get(field) != after_data.get(field)
            ]
            payload = {'changed': changed, 'before': before_data}
            self._dispatch_event(
                self.update_event_name, after_data, payload=payload
            )

        return updated_data

    def create(self, data):
        created_obj = super(CrudManagerWithEvents, self)._create_object(data)
        event_data = self.to_event_serializable(created_obj)
        self._dispatch_event(self.create_event_name, event_data)
        return self.to_serializable(created_obj)

    def delete(self, pk):
        before_obj = self.db_storage.get(pk)
        before_event = self.to_event_serializable(before_obj)
        deleted_data = super(CrudManagerWithEvents, self).delete(pk)
        self._dispatch_event(self.delete_event_name, before_event)
        return deleted_data



from sqlalchemy import inspect
from sqlalchemy_filters import apply_filters, apply_sort


class NotFound(LookupError):
    pass


class DBStorage(object):

    def __init__(self, model_cls, session=None):
        self.model_cls = model_cls
        self.session = session

    def _get(self, pk):
        query = self.query
        # In order to allow the underlying query to be customized with
        # additional filters, we cannot use `query.get` and must construct
        # our own additional PK filter.
        pk_columns = inspect(self.model_cls).primary_key
        pk_values = pk if isinstance(pk, (list, tuple)) else (pk,)

        for col, val in zip(pk_columns, pk_values):
            query = query.filter(getattr(self.model_cls, col.name) == val)

        obj = query.one_or_none()

        if not obj:
            raise NotFound(
                '{} with ID {} does not exist'
                .format(self.model_cls.__name__, pk))
        return obj

    @property
    def query(self):
        return self.session.query(self.model_cls)

    def get(self, pk):
        return self._get(pk)

    def list(self, filters=None, order_by=None, offset=None, limit=None):
        query = self.query
        if filters:
            query = apply_filters(query, filters)
        if order_by:
            query = apply_sort(query, order_by)
        if offset:
            query = query.offset(offset)
        if limit:
            query = query.limit(limit)

        return query.all()

    def count(self, filters=None):
        query = self.query
        if filters:
            query = apply_filters(query, filters)

        return query.count()

    def update(self, pk, data, flush=True, commit=True):
        obj = self._get(pk)
        for key, value in data.items():
            setattr(obj, key, value)
        if commit:
            self.session.commit()
        elif flush:
            self.session.flush()
            self.session.refresh(obj)
        return obj

    def create(self, data, flush=True, commit=True):
        obj = self.model_cls(**data)
        self.session.add(obj)
        if commit:
            self.session.commit()
        elif flush:
            self.session.flush()
            self.session.refresh(obj)

        return obj

    def delete(self, pk, flush=True, commit=True):
        obj = self._get(pk)
        self.session.delete(obj)
        if commit:
            self.session.commit()
        elif flush:
            self.session.flush()



import logging

from nameko.rpc import rpc as nameko_rpc
from nameko.extensions import DependencyProvider

from .managers import CrudManager, CrudManagerWithEvents
from .serializers import default_to_serializable, get_default_from_serializable
from .storage import DBStorage
from .storage import NotFound  # noqa

logger = logging.getLogger(__name__)


def get_dependency_accessor(accessor):

    def get_dependency(service):
        if isinstance(accessor, str):
            return getattr(service, accessor)
        if isinstance(accessor, DependencyProvider):
            # only supported in nameko >= 2.7.0
            return getattr(service, accessor.attr_name)
        # assume a operator.getter style callable:
        return accessor(service)

    return get_dependency


class AutoCrud(DependencyProvider):

    def __init__(
        self, session_provider,
        manager_cls=CrudManager, db_storage_cls=DBStorage,
        model_cls=None,
        from_serializable=None, to_serializable=None,
        get_method_name=None, list_method_name=None,
        page_method_name=None, count_method_name=None,
        create_method_name=None, update_method_name=None,
        delete_method_name=None,
        get_rpc=None, list_rpc=None,
        page_rpc=None, count_rpc=None,
        create_rpc=None, update_rpc=None,
        delete_rpc=None,
        rpc=nameko_rpc,
        **crud_manager_kwargs
    ):
        required = [
            (session_provider, 'session_provider'),
            (manager_cls, 'manager_cls'),
            (db_storage_cls, 'db_storage_cls'),
            (model_cls, 'model_cls'),
        ]
        missing = [name for param, name in required if not param]
        if missing:
            raise ValueError(
                '`{}` param(s) are missing for {}'.format(
                    missing, type(self).__name__))

        # store these providers as a map so they are not seen by nameko
        # as sub-dependencies
        self.session_accessor = get_dependency_accessor(session_provider)
        self.model_cls = model_cls
        self.manager_cls = manager_cls
        self.db_storage_cls = db_storage_cls
        self.crud_manager_kwargs = crud_manager_kwargs
        self.rpc = rpc

        self.method_config = {
            'get': (get_method_name, get_rpc),
            'list': (list_method_name, list_rpc),
            'page': (page_method_name, page_rpc),
            'count': (count_method_name, count_rpc),
            'create': (create_method_name, create_rpc),
            'update': (update_method_name, update_rpc),
            'delete': (delete_method_name, delete_rpc),
        }

        self.from_serializable = (
            from_serializable or get_default_from_serializable(model_cls))

        self.to_serializable = (to_serializable or default_to_serializable)

    def bind(self, container, attr_name):
        """
        At bind time, modify the service class to add additional rpc methods.
        """
        service_cls = container.service_cls

        bound = super(AutoCrud, self).bind(container, attr_name)

        def make_manager_fn(fn_name):
            def _fn(self, *args, **kwargs):
                """ This is the RPC method that will run on the service """
                # instantiate a manager instance
                manager = bound.manager_cls(
                    bound,  # the provider
                    self,  # the service instance
                    db_storage=getattr(self, attr_name),
                    from_serializable=bound.from_serializable,
                    to_serializable=bound.to_serializable,
                    **bound.crud_manager_kwargs
                )
                # delegate to the manager method with the same name.
                return getattr(manager, fn_name)(*args, **kwargs)
            return _fn

        for mngr_fn_name, (rpc_name, method_rpc) in self.method_config.items():
            if rpc_name and not getattr(service_cls, rpc_name, None):
                manager_fn = make_manager_fn(mngr_fn_name)
                setattr(service_cls, rpc_name, manager_fn)
                # apply rpc decorator
                rpc = method_rpc or self.rpc
                rpc(manager_fn)

        return bound

    def get_dependency(self, worker_ctx):
        # returns a storage instance without session
        # session is bound to it at worker_setup
        return self.db_storage_cls(self.model_cls)

    def worker_setup(self, worker_ctx):
        service = worker_ctx.service
        session = self.session_accessor(service)

        # add required session to the storage
        db_storage = getattr(service, self.attr_name)
        db_storage.session = session


class AutoCrudWithEvents(AutoCrud):

    def __init__(
        self,
        session_provider,
        dispatcher_provider,
        event_entity_name,
        create_event_name=None,
        update_event_name=None,
        delete_event_name=None,
        manager_cls=CrudManagerWithEvents,
        **kwargs
    ):
        required = [
            (dispatcher_provider, 'dispatcher_provider'),
            (event_entity_name, 'event_entity_name'),
        ]
        missing = [name for param, name in required if not param]
        if missing:
            raise ValueError(
                '`{}` param(s) are missing for {}'.format(
                    missing, type(self).__name__))

        dispatcher_accessor = get_dependency_accessor(dispatcher_provider)
        super(AutoCrudWithEvents, self).__init__(
            session_provider,
            manager_cls=manager_cls,
            dispatcher_accessor=dispatcher_accessor,
            event_entity_name=event_entity_name,
            create_event_name=create_event_name,
            update_event_name=update_event_name,
            delete_event_name=delete_event_name,
            **kwargs
        )



from datetime import date, datetime
from decimal import Decimal

from dateutil import parser
from enum import Enum

import sqlalchemy as sa
from sqlalchemy_utils import ChoiceType


def default_to_serializable(obj):
    """ Convert a sqlalchemy model instance to a dict ready for serialization.
    """
    try:
        dict_ = obj.to_dict()
    except AttributeError:
        dict_ = {
            col.name: getattr(obj, col.name)
            for col in obj.__table__.columns
        }

    def get_value(val):
        if val is None:
            return None
        if isinstance(val, (str, int, float, bool, list, dict)):
            return val
        # TODO- can't use Enum in py2
        if isinstance(val, Enum):
            return val.value
        if isinstance(val, (date, datetime)):
            return val.isoformat()
        return str(val)

    return {field: get_value(val) for field, val in dict_.items()}


def get_default_from_serializable(model_cls):
    col_dict = {
        col.name: col
        for col in model_cls.__table__.columns
    }

    def default_from_serializable(dict_):
        """ Convert serialized field-values dict into values that can be
            supplied to a sqlalchemy object
        """
        def get_value(field, val):
            if val is None:
                return None

            col = col_dict[field]
            if isinstance(col.type, (sa.Date, sa.DateTime)):
                return parser.parse(val)
            if isinstance(col.type, sa.DECIMAL):
                return Decimal(val)
            if isinstance(col.type, ChoiceType):
                return col.type.type_impl.process_result_value(val, None)

            return val

        return {field: get_value(field, val) for field, val in dict_.items()}

    return default_from_serializable



from setuptools import setup, find_packages

setup(
    name='nameko-autocrud',
    version='0.2.0',
    description='Autocrud utility for nameko services',
    author='timbu',
    url='http://github.com/timbu/nameko-autocrud',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=[
        "nameko>=2.6.0",
        "sqlalchemy>=1.0.16",
        "sqlalchemy_filters>=0.4.0",
        "python-dateutil>=2.6.1",
        "sqlalchemy-utils>=0.32.5",
    ],
    extras_require={
        'dev': [
            "coverage==4.0.3",
            "flake8==2.5.4",
            "pylint==1.5.5",
            "pytest==3.0.5",
            "nameko_sqlalchemy>=0.1.0",
        ]
    },
    zip_safe=True,
    license='Apache License, Version 2.0',
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ]
)
